MODULE fnn;
(*
 *
 *      FALSE NEAREST NEIGHBORS CODE
 *
 *
 *      Usage:
 *
 *      fnn -d lowdim highdim -N Npoints -t time-delayembedding
 *          -B datafile -r RTOL
 *
 *  Matt Kennel.  5/31/92.  This program is in the public domain, and
 *  has no restrictions on it, except that you cite the method used in
 *  any publication or technical report.   This relinquishing of copyright
 *  doesn't apply to 
 *)

FROM    DynArray        IMPORT
    MakeArray,CardArrayPtr,RealArrayPtr;

FROM    DynMatrix       IMPORT
    MakeMatrix, RealMatrixPtr;

FROM    Util            IMPORT
    ScanInteger,UniformRandom, ArgvStringExists, FatalError, Exit;

FROM    KdTree          IMPORT
    KdTreePtr,CreateKdTree,SearchKdTree;

FROM	SYSTEM		IMPORT
    LN, EXP, SQRT;

IMPORT	IO;


VAR

    OneDData        :   DoubleArrayPtr;   (* The 1-D data array *)
    DoLPS	    :	BOOLEAN;	(* do we run LPS? *)
    Ndata           :   CARDINAL;       (* the number of data points *)

    Nvectors        :   CARDINAL;       (* The number of vectors that we can
					    form given the dimension and time
					    delay *)

    DimLo, DimHi    :   CARDINAL;       (* range of dimensions to loop over *)
    Tau             :   CARDINAL;       (* time delay, in units of sampling
					    time*)

    Rtol	    :	LONGREAL;

    Size	    :	LONGREAL;	(* typical size of the attractor *)

    d		    :	CARDINAL;

PROCEDURE   ReadInData;
(*
 *    From the arguments passed to the program get the input data vector
 *    Either from an ascii file, or from a binary file:
 *
 *    fnn -A filename.ascii
 *   or
 *    fnn -B filename.binary
 *
 *    -B is searched for FIRST on the command line
 *    Allocates OneDData;
 *)

VAR
    argplace        :   CARDINAL;
    binfile	    :	FILE;
    asciifile	    :	TEXT;
    i		    :	CARDINAL;
BEGIN

    argplace := ArgvStringExists("-N");
    IF (argplace = 0) THEN
	FatalError("You must provide a -N npoints flag!");
    END;

    Ndata := ScanInteger(PARAMSTR(argplace+1));
    writeln('Ndata = ',Ndata);

    OneDData := MakeDoubleArray(Ndata);
					(*allocate the array *)


    argplace := ArgvStringExists("-B");
    IF (argplace > 0) THEN              (* found binary file *)
	RESET(binfile,PARAMSTR(argplace+1));
	BLOCKREAD(binfile,OneDData^,TSIZE(LONGREAL)*Ndata);
    ELSE
	argplace := ArgvStringExists("-A");
	IF (argplace > 0) THEN
	    RESET(asciifile,PARAMSTR(argplace+1));
	    FOR i := 0 TO Ndata-1 DO
		READ(asciifile,OneDData^[i]);
	    END;
	ELSE
	    FatalError("You must provide either a -A (ascii) or -B (binary) file");
	END;
    END;

END ReadInData;


PROCEDURE GetOptions;
(*
 *  Get the other pieces of data from the command line :
 *
 *  -d dim1 dim2
 *)
VAR
    argplace	    :	CARDINAL;

BEGIN
    IF (PARAMCOUNT = 0) THEN
	IO.WritelnS(IO.StdError,"usage:");
	IO.WritelnS(IO.StdError,"fnn -N npoints -d dim1 dim2 -t tau -r rtol");
	IO.WritelnS(IO.StdError,"{-LPS} {-B datafile.dbin}|{-A datafile.ascii}");

	Exit(0);
    END;

    argplace := ArgvStringExists("-d");
    IF (argplace = 0) THEN
	FatalError("You must provide a -d flag!!!");
    END;
    DimLo := ScanInteger(PARAMSTR(argplace+1));
    DimHi := ScanInteger(PARAMSTR(argplace+2));

    ASSERT(DimLo < DimHi);

    argplace := ArgvStringExists("-t");
    IF (argplace = 0) THEN
	FatalError("You must provide a -t flag!!!");
    END;
    Tau := ScanInteger(PARAMSTR(argplace+1));

    argplace := ArgvStringExists("-r");
    IF (argplace = 0) THEN
	FatalError("You must provide a -r flag!!!");
    END;
    Rtol := ScanLongReal(PARAMSTR(argplace+1));

    argplace := ArgvStringExists("-LPS");
    IF (argplace = 0) THEN
	DoLPS := FALSE;
    ELSE
	DoLPS := TRUE;
    END;
    writeln('DimLo:',DimLo:5,' DimHi:',DimHi:5,' Tau:',Tau,' Rtol:',Rtol:6:2);
    

END GetOptions;

PROCEDURE   EmbedData(oned:	DoubleArrayPtr;
		      ndata:	CARDINAL;
		      d:	CARDINAL;
		      tau:	CARDINAL;
		      VAR nvec:	CARDINAL ) : RealMatrixPtr;
(* 
 *  Do the time-delay embedding, dynamically creating a "RealMatrixPtr",
 *  size Length nvec, width d.
 *  returning the number of valid vectors into nvec.  
 *  the RealMatrixPtr is accessed as      m^[0..d-1]^[0..nvec-1]
 *)
VAR
    i,j	    :	    CARDINAL;
    m	    :	    RealMatrixPtr;

BEGIN

    nvec := ndata - (d * tau);
    m := MakeMatrix(nvec, d);

    FOR i := 0 TO nvec-1 DO
	FOR j := 0 TO d-1 DO
	    m^[i]^[j] := oned^[i + j*tau];
	END;
    END;
    RETURN(m);

END EmbedData;

PROCEDURE DoSizeOfAttractor;

VAR
    i	    :	CARDINAL;
    m,d,tmp :	LONGREAL;
    
BEGIN

    m := 0.0;
    FOR i := 0 TO Ndata-1 DO
	m := m + OneDData^[i];
    END;
    m := m / Ndata;

    d := 0.0;
    FOR i := 0 TO Ndata-1 DO
	d := d + (OneDData^[i]-m) ** 2;
    END;

    Size := SQRT(d / Ndata);
    
END DoSizeOfAttractor;



PROCEDURE   CalculateFalseNeighbors(d : CARDINAL);
(*
 * actually calculate the false neighbors.
 *)
VAR
    embeddata	:	RealMatrixPtr;
    tree	:	KdTreePtr;
    i,j		:	CARDINAL;
    nfalse	:	CARDINAL;
    indexes	:	CardArrayPtr;
    distances	:	RealArrayPtr;
    r1,rext	:	LONGREAL;

BEGIN
    writeln('Embedding data, for d = ',d:4);
    embeddata := EmbedData(OneDData, Ndata, d, Tau, Nvectors);

    writeln('Creating Tree: Nvectors = ', Nvectors,' d = ',d);
    tree := CreateKdTree(embeddata, Nvectors, d);

    Nvectors := Nvectors - 1;

    indexes := MakeCardArray(10);
    distances := MakeRealArray(10);

    nfalse := 0;
    FOR i := 0 TO Nvectors-1 DO
	(* find the nearest neighbor! *)

	SearchKdTree(tree, embeddata^[i], 2, indexes, distances);
	(* writeln('Neighbor index = ',indexes^[1]:6,'at dis ',distances^[1]:8); *)

	r1 := SQRT(distances^[1]);
	rext := OneDData^[i + d*Tau] - OneDData^[indexes^[1] + d*Tau];
	IF (rext < 0) THEN
	    rext := -rext;
	END;
	(* rext is the extra distance incurred for going to the next highest dimension *)

	IF (rext / r1) > Rtol THEN   (* we have a false neighbor *)
	    nfalse := nfalse + 1;
	ELSIF (rext / Size) > 2.0 THEN  (* second test *)
	    nfalse := nfalse + 1;
	END;

	IF (i MOD 100) = 0 THEN
	    write('.');
(*EMBED
 fflush(stdout);
 *)
	END;

    END;

    writeln;
    writeln('Proportion of f.n.n. = ',nfalse:6,'/',Nvectors:6, (nfalse/Nvectors):10:3);

    writeln('Destroying Tree...');
    DestroyKdTree(tree);
    FreeMatrix( embeddata, Nvectors, d);

END CalculateFalseNeighbors;
 
PROCEDURE EuclideanDistance(index1: CARDINAL;
			    index2: CARDINAL;
			    d	  : CARDINAL) : LONGREAL;

VAR
    dis,tmp : LONGREAL;
    i	    : CARDINAL;

BEGIN
    dis := 0.0;
    FOR i := 0 TO d DO
	tmp := OneDData^[index1 + i*Tau] - OneDData^[index2 + i*Tau];
	dis := dis + tmp*tmp;
    END;

    RETURN(SQRT(dis));

END;

PROCEDURE TwoEuclideanDistances(index1	: CARDINAL;
				index2	: CARDINAL;
				d	: CARDINAL;
				VAR outd: LONGREAL;
				VAR outdp1: LONGREAL);

VAR
    dis,tmp : LONGREAL;
    i	    : CARDINAL;

BEGIN
    dis := 0.0;
    FOR i := 0 TO d-1 DO
	tmp := OneDData^[index1 + i*Tau] - OneDData^[index2 + i*Tau];
	dis := dis + tmp*tmp;
    END;

    outd := SQRT(dis);

    tmp := OneDData^[index1 + d*Tau] - OneDData^[index2 + d*Tau];
    dis := dis + tmp*tmp;

    outdp1 := SQRT(dis);

END;


PROCEDURE   CalculateLPS(d : CARDINAL);
(*
 * calculate the LPS statistic.
 *)
CONST
    nneigh	=	1;

VAR
    embeddata1	:	RealMatrixPtr;
    embeddata2	:	RealMatrixPtr;
    tree1	:	KdTreePtr;
    tree2	:	KdTreePtr;
    i,j,k	:	CARDINAL;
    nfalse	:	CARDINAL;
    indexes1	:	CardArrayPtr;
    indexes2	:	CardArrayPtr;
    distances1	:	RealArrayPtr;
    distances2	:	RealArrayPtr;
    r1,rext	:	LONGREAL;
    a1,a2,a3,a4	:	LONGREAL;
    Q1, Q2	:	LONGREAL;
    W,W2	:	LONGREAL;

BEGIN
    writeln('Embedding data, for d = ',d:4);
    embeddata1 := EmbedData(OneDData, Ndata, d, Tau, Nvectors);

    writeln('Creating First Tree...');
    tree1 := CreateKdTree(embeddata1, Nvectors, d);

    writeln('Embedding data, for d = ',(d+1):4);
    embeddata2 := EmbedData(OneDData, Ndata, d+1, Tau, Nvectors);

    writeln('Creating Second Tree...');
    tree2 := CreateKdTree(embeddata2, Nvectors, d+1);



    Nvectors := Nvectors - 1;

    indexes1 := MakeCardArray(20);
    distances1 := MakeRealArray(20);
    indexes2 := MakeCardArray(20);
    distances2 := MakeRealArray(20);

    nfalse := 0;

    W := 0.0;

    FOR i := 0 TO Nvectors-1 DO
	(* find the nearest neighbor! *)

	SearchKdTree(tree1, embeddata1^[i], nneigh+1, indexes1, distances1);
	SearchKdTree(tree2, embeddata2^[i], nneigh+1, indexes2, distances2);
	(* writeln('Neighbor index = ',indexes^[1]:6,'at dis ',distances^[1]:8); *)

	W2 := 1.0;
	FOR j := 1 TO nneigh DO

(* Calculate distances *)
(*	    Q1 := EuclideanDistance(i, indexes1^[j], d+1) /
		  EuclideanDistance(i, indexes2^[j], d+1);

	    Q2 := EuclideanDistance(i, indexes1^[j], d)/
		  EuclideanDistance(i, indexes2^[j], d);
 *)

	    TwoEuclideanDistances(i, indexes1^[j],d, a3, a1);
	    TwoEuclideanDistances(i, indexes2^[j],d, a4, a2);
	    Q1 := a1/a2;
	    Q2 := a3/a4;

	    
	    W2 := W2  *Q1*Q2 ;
	END;

(*	writeln('Local W = ',EXP( LN(W2) / (2*nneigh) ) ); *)
	W := W + EXP( LN(W2) / (2*nneigh) );
	(* take the geometric mean *)
    
	IF (i MOD 100) = 0 THEN
	    write('.');
(*EMBED
 fflush(stdout);
 *)
	END;

    END;


    W := LN(W / (Nvectors));

    writeln;
    writeln('d, W, W/T = ',d:5,' ',W:10,' ',W/Tau:10);

    writeln('Destroying Tree...');
    DestroyKdTree(tree1);
    FreeMatrix(embeddata1, Nvectors, d);

    DestroyKdTree(tree2);
    FreeMatrix(embeddata2, Nvectors, d+1);

END CalculateLPS;
 

BEGIN                                   (* the main program *)

    GetOptions;
    ReadInData;
    DoSizeOfAttractor;
    
    FOR d := DimLo TO DimHi DO

	CalculateFalseNeighbors(d);
	IF DoLPS THEN
	    CalculateLPS(d);
	END;

    END;

END fnn.
